print("import: 'rdflib'")
import rdflib

print("import: 'rdflib.extras'")
import rdflib.extras

print("import: 'rdflib.plugins'")
import rdflib.plugins

print("import: 'rdflib.plugins.parsers'")
import rdflib.plugins.parsers

print("import: 'rdflib.plugins.parsers'")
import rdflib.plugins.parsers

print("import: 'rdflib.plugins.serializers'")
import rdflib.plugins.serializers

print("import: 'rdflib.plugins.sparql'")
import rdflib.plugins.sparql

print("import: 'rdflib.plugins.stores'")
import rdflib.plugins.stores

print("import: 'rdflib.tools'")
import rdflib.tools

